#include <math.h>
#include <robotTypes.h>
#include "accTrack.h"

static void compRadTrack (
double *newp,
double *newv,
double p0,
double v0,
double pt,
MotionContext *ctx)
{
	double e0;		/* initial energy */
	double emax;		/* peak energy */
	double vmax;		/* peak velocity */
	double ad;		/* final acceleration towards target */
	double px;		/* intersection of accel and decel curves */
	double t1, tc, t2;	/* accel, cruise, and deccel times */
	double tr;		/* time inside accel, cruise, or deccel */
	double pn, en, vn;	/* new position, energy, and velocity values*/
	double tau = 1/ctx->rate;/* sample time */

	e0 = -v0*v0/2;
	/* set ad -/+ depending on whether we expect to overshoot */
	ad = (v0 > 0 && e0 < ctx->accMax*(p0-pt)) ? -ctx->accMax : ctx->accMax;

	px = (pt + p0 + e0/ad)/2;
	emax = (ad*(p0 - pt) + e0)/2;

	tc = 0.0;			/* tc and vmax are changed */
	vmax = SGN(ad)*sqrt(-2*emax);	/*  if velocity is clipped */
	if (SGN(ad) == 1)
	 { /* check for velocity limits if we are not going to overshoot */
	   if (emax < -ctx->velMax*ctx->velMax/2)
	    { vmax = ctx->velMax;
	      emax = -ctx->velMax*ctx->velMax/2;
	      tc = ((2*emax-e0)/ad + (pt-p0))/vmax;
	    }
	 }
	t1 = (vmax-v0)/ad;
	t2 = vmax/ad;
	if (tau >= t1 + tc + t2)
	 { vn = 0.0;
	   pn = pt;
	 }
	else if ((tr = tau - (t1+tc)) > 0)
	 { vn = vmax-ad*tr;
	   pn = -vn*vn/(2*ad) + pt;
	 }
	else if ((tr = tau - t1) > 0)
	 { vn = vmax;
	   pn = p0 + (e0-emax)/ad + tr*vn;
	 }
	else
	 { vn = v0 + ad*tau;
	   pn = p0 + v0*tau + ad*tau*tau/2;
	 }
	*newp = pn;
	*newv = vn;
}

#define EPSILON	(1e-6)

accTrack(new, pd, old, ctx)
MotionState *new;
VECT *pd;
MotionState *old;
MotionContext *ctx;
{
	VECT Dp0;	/* vector to target */
	VECT Dp0_unit;	/* radial unit vector to target */
	VECT Vr0;	/* projection of VelOld onto Dp0 */
	double vr0_mag;	/* magnitude of projection of VelOld onto Dp0 */
	VECT Vp0;	/* part of VelOld perpendicular to Dp0 */
	VECT tmp;

	double dp0_mag;	/* ||Dp0|| */
	double newp, newv;

/*	Dp0.z = Dp0_unit.z = Vr0.z = Vp0.z = tmp.z = 0.0; */
	
	subtractVect (&Dp0, pd, &old->pos);
	dp0_mag = normVect (&Dp0);
	if (dp0_mag < EPSILON)
	 { dp0_mag = 0;
	   if (normVect(&old->vel) < EPSILON)
	    { new->pos = *pd;
	      zeroVect (&new->vel);
	      return (1);
	    }
	   else
	    { unitVect (&Dp0_unit, &old->vel);
	      Vr0 = old->vel;
	      zeroVect (&Vp0);
	    }
	 }
	else
	 { scaleVect (&Dp0_unit, 1/dp0_mag, &Dp0);
	   vr0_mag = dotVect (&Dp0_unit, &old->vel);
	   scaleVect (&Vr0, vr0_mag, &Dp0_unit);
	   subtractVect (&Vp0, &old->vel, &Vr0);
	 }
	
	/* set the radial velocity and position */

/*
	accTrack1Pos() was put in here for test purposes. It should
	behave identically to compRadTrack().

	accTrack1Pos (&newp, &newv, -dp0_mag, vr0_mag, 
		      ctx->velMax, ctx->accMax, 1/ctx->rate);
*/

	compRadTrack (&newp, &newv, -dp0_mag, vr0_mag, 0, ctx);
	scaleVect (&tmp, newp, &Dp0_unit);
	addVect (&new->pos, pd, &tmp);
	scaleVect (&new->vel, newv, &Dp0_unit);

	if (normVect(&Vp0) > EPSILON)
	 { 
	   compRadTrack (&newp, &newv, 0, normVect(&Vp0), 0, ctx);
	   unitVect (&tmp, &Vp0);
	   scaleVect (&tmp, newp, &tmp);
	   addVect (&new->pos, &new->pos, &tmp);
	   unitVect (&tmp, &Vp0);
	   scaleVect (&tmp, newv, &tmp);
	   addVect (&new->vel, &new->vel, &tmp);
	 }
	return (0);
}
