package acme.server.platform.PanTilt;

import java.io.*;

/** An implementation of control for PanTilt.
 *
 * @author Lun Ke
 * @version %I %G%
 */
public class PanTiltDevice {

  private RandomAccessFile port;

  private final int BUFLEN = 1024; 
  private final String portName = "com2";
  private final byte STX = 0x02;
  private final byte ETX = 0x03;
  private final byte Address = 0x00;

  /**
   * Constructor
   */
  public PanTiltDevice() throws IOException {
    port = new RandomAccessFile(portName, "rw");
  }

  /**
   * Initializes the PanTilt. 
   * Set default PanTilt prameters. 
   */
  public boolean initialize() {

    //Set command set to be shortb.
    setCommandSet( (byte) 0x00 );
    //Set coordinate system to be upside down.
    setCoordinate( (byte) 0x01 );
    //Set the direction and speed.
    setDirectionSpeed( (byte) 0xff, (byte) 0x10, (byte) 0x00,
                       (byte) 0x10, (byte) 0x00 );
   
    return( true );
  }

  /** 
   * Generate the CRC value for the string. 
   */ 
  private static byte CRC(byte[] data, int size) {
    int offset;
    int CRC_val = 0x00;

    for(offset = 0; offset < size; offset++) {
      CRC_val = (CRC_val>>4)^(((CRC_val^data[offset]) & 0x0f) * 0x1081);
      CRC_val = (CRC_val>>4)^(((CRC_val^(data[offset]>>4)) & 0x0f) * 0x1081);
    }

    return (byte)( CRC_val & 0x000000ff );
  }

  /**
   * Get the OS version of PanTilt.
   */
  public String getVersion() {
    int size = 0;
    byte[] buf = new byte[BUFLEN];

    buf[0] = STX;
    buf[1] = Address;
    buf[2] = 0x00;	//command byte
    buf[3] = PanTiltDevice.CRC(buf, 3);
    buf[4] = ETX;

    try {
      port.write(buf, 0, 5);
    } catch (IOException e) {
      System.err.println("getVersion:Could not write to serial port for PanTilt.");
      System.exit(1);
    }

    try {
      port.read(buf, 0, 4);
      if( buf[2] != 0 ) { // Error bit is non-zero
        System.out.println("getVersion:Error Occured in execution.");
        System.exit(1);
      }
      size = buf[3];
      port.read(buf, 4, size+2);
    } catch (IOException e) {
      System.err.println("getVersion:Could not read result from Pantilt.");
      System.exit(1);
    }

    return( new String(buf, 0, size) );
  }

  /**
   * Set the command set for PanTilt.
   * flag: 0    short command set
   *       1    long command set
   */
  public boolean setCommandSet(byte flag) {
    byte[] buf = new byte[BUFLEN];

    buf[0] = STX;
    buf[1] = Address;
    buf[2] = (byte)0x82;	//command byte
    buf[3] = flag;
    buf[4] = PanTiltDevice.CRC(buf, 4);
    buf[5] = ETX;

    try {
      port.write(buf, 0, 6);
    } catch (IOException e) {
      System.err.println("setCommandSet:Could not write to serial port for PanTilt.");
      System.exit(1);
    }

    try {
      port.read(buf, 0, 5);
      if( buf[2] != 0 ) { // Error bit is non-zero
        System.out.println("setCommandSet:Error Occured in execution.");
        System.exit(1);
      }
    } catch (IOException e) {
      System.err.println("setCommandSet:Could not read result from Pantilt.");
      System.exit(1);
    }

    return( true );
    //System.out.write( buf, 0, 100 );
  }
  
  /**
   * Set coordinate system
   * flag: 0x00    tabletop
   *       0x04    sideways on a wall
   *       0x01    upside down
   */
  public boolean setCoordinate(byte flag) {
    byte[] buf = new byte[BUFLEN];

    buf[0] = STX;
    buf[1] = Address;
    buf[2] = 0x3f;	//command byte
    buf[3] = flag;
    buf[4] = PanTiltDevice.CRC(buf, 4);
    buf[5] = ETX;

    try {
      port.write(buf, 0, 6);
    } catch (IOException e) {
      System.err.println("setCoordinate:Could not write to serial port for PanTilt.");
      System.exit(1);
    }

    try {
      port.read(buf, 0, 5);
      if( buf[2] != 0 ) { // Error bit is non-zero
        System.out.println("setCoordinate:Error Occured in execution.");
        System.exit(1);
      }
    } catch (IOException e) {
      System.err.println("setCoordinate:Could not read result from Pantilt.");
      System.exit(1);
    }

    return( true );
  }
   
  /**
   * Set the direction and speed of the azimuth and elevation movement.
   * direction: bit0, elevation direction, 0-negative, 1-positive
   *            bit1, elevation motion,    0-stop,     1-move
   *            bit2, azimuth direction,   0-negative, 1-positive
   *            bit3, azimuth motion,      0-stop,     1-move
   *            0x00ff for positive direction and movable;
   * speed: 4 bytes. the first 2 stand for azimuth speed,
   *        the second 2 stand for elevation speed.
   */
  public boolean setDirectionSpeed(byte direction, byte aspeed0, byte aspeed1,
				   byte espeed0, byte espeed1) {
    byte[] buf = new byte[BUFLEN]; 

    buf[0] = STX;
    buf[1] = Address;
    buf[2] = 0x18;	//command byte
    buf[3] = direction;
    buf[4] = aspeed0;
    buf[5] = aspeed1;
    buf[6] = espeed0;
    buf[7] = espeed1;
    buf[8] = PanTiltDevice.CRC(buf, 8);
    buf[9] = ETX;

    try {
      port.write(buf, 0, 10);
    } catch (IOException e) {
      System.err.println("setDirectionSpeed:Could not write to serial port for PanTilt.");
      System.exit(1);
    }

    try {
      port.read(buf, 0, 5);
      if( buf[2] != 0 ) { // Error bit is non-zero
        System.out.println("setDirectionSpeed:Error Occured in execution.");
        System.exit(1);
      }
    } catch (IOException e) {
      System.err.println("setDirectionSpeed:Could not read result from Pantilt.");
      System.exit(1);
    }

    return( true );
  }

  /**
   * Set the PanTilt's orientation by azimuth and elevation degree.
   */
  public boolean setOrientation(float azimuth, float elevation) {
    //Convert degree to number of steps( 0.01 degree/step )
    int stepa = (int)(azimuth*100);
    int stepe = (int)(elevation*100);

    byte stepal = (byte)(stepa & 0xff); 
    byte stepah = (byte)((stepa>>8) & 0xff);

    byte stepel = (byte)(stepe & 0xff); 
    byte stepeh = (byte)((stepe>>8) & 0xff);

    return(setOrientationCmd(stepah, stepal, stepeh, stepel));
  }
  
  /**
   * Set the PanTilt's orientation by PanTilt's command.
   */
  public boolean setOrientationCmd(byte apos0, byte apos1, byte epos0, byte epos1) {
    byte[] buf = new byte[BUFLEN];

    buf[0] = STX;
    buf[1] = Address;
    buf[2] = 0x19;      //command byte
    buf[3] = apos0;
    buf[4] = apos1;
    buf[5] = epos0;
    buf[6] = epos1;
    buf[7] = PanTiltDevice.CRC(buf, 7);
    buf[8] = ETX;

    try {
      port.write(buf, 0, 9);
    } catch (IOException e) {
      System.err.println("setOrientation:Could not write to serial port for PanTilt.");
      System.exit(1);
    }

    try {
      port.read(buf, 0, 5);
      if( buf[2] != 0 ) { // Error bit is non-zero
        System.out.println("setOrientation:Error Occured in execution.");
        System.exit(1);
      }
    } catch (IOException e) {
      System.err.println("setOrientation:Could not read result from Pantilt."
);
      System.exit(1);
    }

    return( true );
  }

  /**
   * Stop the movement of PanTilt.
   */
  public boolean stopMotion() {
    byte[] buf = new byte[BUFLEN];

    buf[0] = STX;
    buf[1] = Address;
    buf[2] = 0x14;	//command byte
    buf[3] = PanTiltDevice.CRC(buf, 3);
    buf[4] = ETX;

    try {
      port.write(buf, 0, 5);
    } catch (IOException e) {
      System.err.println("stopMotion:Could not write to serial port for PanTilt.");
      System.exit(1);
    }

    try {
      port.read(buf, 0, 4);
      if( buf[2] != 0 ) { // Error bit is non-zero
        System.out.println("stopMotion:Error Occured in execution.");
        System.exit(1);
      }
    } catch (IOException e) {
      System.err.println("stopMotion:Could not read result from Pantilt.");
      System.exit(1);
    }

    return( true );
  }
    
}

