/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd and Vincent Hayward.	    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*=~=~=~=~ McGill-Computer-Vision-and-Robotics-Laboratory--C-Module ~=~=~=~=
 *		
 * MODULE NAME:	gantryJacob.c
 *
 * DESCRIPTION:	Jacobian computations for the Stanford arm.
 *		For description, see 'man pumaJacob'
 * 
 * AUTHOR/DATE:	John Lloyd, Spring 1987, after Vincent Hayward, Purdue, 1983
 *		Rewrote routines to interface with new RCI, winter 1988.
 *
 *~=~=~=~=~=~= McGill-Computer-Vision-and-Robotics-Laboratory ~=~=~=~=~=~=*/

#include <math.h>

#include "../h/fastmath.h"
#include "../h/cdefs.h"
#include "../h/robot_jls.h"
#include "../h/robotMath.h"
#include "../h/gantry_jls.h"
#include "../h/gantry_kynvar.h"
#include "../h/robotTypes.h"

#define EPSILONF 2.0e-6

#define J1	0
#define J2	1
#define J3	2
#define J4	3
#define J5	4

/* ---- Definition of var structure parameters ---- */

#define S4	svar->s4
#define C4	svar->c4
#define S5	svar->s5
#define C5	svar->c5

gantryFwdJacobT(t, f, j6, kyn)
float *t;
FORCE *f;
float *j6;
KYN *kyn;
{
	GANTRY_VAR var;

	gantryUpdateVar (&var, j6, kyn);
	return (gantryFwdJacobTVar (t, f, &var, kyn));
}

gantryFwdJacob(dc, dj, j6, kyn)
DIFF *dc;
float *dj;
float *j6;
KYN *kyn;
{
	GANTRY_VAR var;

	gantryUpdateVar (&var, j6, kyn);
	return (gantryFwdJacobVar (dc, dj, &var, kyn));
}

gantryInvJacobT(f, t, j6, kyn, epsilon)
FORCE *f;
float *t;
float *j6;
KYN *kyn;
double epsilon;
{
	GANTRY_VAR var;

	gantryUpdateVar (&var, j6, kyn);
	return (gantryInvJacobTVar (f, t, &var, kyn, epsilon));
}

gantryInvJacob(dj, dc, j6, kyn, epsilon)
float *dj;
DIFF *dc;
float *j6;
KYN *kyn;
double epsilon;
{
	GANTRY_VAR var;

	gantryUpdateVar (&var, j6, kyn);
	return (gantryInvJacobVar (dj, dc, &var, kyn, epsilon));
}

gantryFwdJacobTVar(j, f, var, kyn) /*::*/
float *j;
FORCE *f;
void *var;
KYN *kyn;
{
	register GANTRY_VAR *svar = (GANTRY_VAR*)var;

	j[J1] = -C4*C5*f->f.x + S4*f->f.y - C4*S5*f->f.z;
	j[J2] =  C5*S4*f->f.x + C4*f->f.y + S4*S5*f->f.z;
	j[J3] =     S5*f->f.x                 -C5*f->f.z;
	j[J4] =    -S5*f->m.x +                C5*f->m.z;
	j[J5] =                    f->m.y;
}

gantryFwdJacobVar(d, q, var, kyn) /*::*/
register DIFF_PTR d;
register float *q;
void *var;
KYN *kyn;
{
	register GANTRY_VAR *svar = (GANTRY_VAR*)var;

	d->t.x = -C4*C5*q[J1] + C5*S4*q[J2] + S5*q[J3];
	d->t.y =     S4*q[J1] +    C4*q[J2];
	d->t.z = -C4*S5*q[J1] + S4*S5*q[J2] - C5*q[J3];
	d->r.x =    -S5*q[J4];
	d->r.y =                      q[J5];
	d->r.z =     C5*q[J4];
}

gantryInvJacobVar(q, d, var, kyn, epsilon)
register float *q;
register DIFF_PTR d;
void *var;
KYN *kyn;
double epsilon;
{
	GANTRY_KYN *skyn = (GANTRY_KYN*)kyn->ext;
	register GANTRY_VAR *svar = (GANTRY_VAR*)var;
	int code = 0;

	if (epsilon < EPSILONF)
	 { epsilon = EPSILONF;
	 }
	q[J1] = -C4*C5*d->t.x + S4*d->t.y - C4*S5*d->t.z;
	q[J2] =  C5*S4*d->t.x + C4*d->t.y + S4*S5*d->t.z;
	q[J3] =     S5*d->t.x                 -C5*d->t.z;
   	if (FABS(S5) < epsilon)
	 { code |= GANTRY_WRIST_DEG;
	   q[J4] = -d->r.x*S5 / (epsilon*epsilon);
	 }
	else
	 { q[J4] = -d->r.x/S5;
	 }
	q[J5] = d->r.y;
	return (code);
}

gantryInvJacobTVar(f, t, var, kyn, epsilon)
register FORCE *f;
register float *t;
void *var;
KYN *kyn;
double epsilon;
{
	GANTRY_KYN *skyn = (GANTRY_KYN*)kyn->ext;
	register GANTRY_VAR *svar = (GANTRY_VAR*)var;
	int code = 0;

	/* Set threshold to something reasonable */

	if (epsilon < EPSILONF)
	 { epsilon = EPSILONF;
	 }
	f->f.x = -C4*C5*t[J1] + C5*S4*t[J2] + S5*t[J3];
	f->f.y =     S4*t[J1] +    C4*t[J2];
	f->f.z = -C4*S5*t[J1] + S4*S5*t[J2] - C5*t[J3];
   	if (FABS(S5) < epsilon)
	 { code |= GANTRY_WRIST_DEG;
	   f->m.x = -t[J4]*S5 / (epsilon*epsilon);
	 }
	else
	 { f->m.x = -t[J4]/S5;
	 }
	f->m.y = t[J5];
	f->m.z = 0;
	return (code);
}

/* Jacobian updating functions */

gantryUpdateVar (var, ang, kyn)
void *var;
register float *ang;
KYN *kyn;
{
	register GANTRY_VAR *svar = (GANTRY_VAR*)var;

	SINCOS (ang[J4], &S4, &C4);
	SINCOS (ang[J5], &S5, &C5);

	gantryUpdateVarXsincos (var, ang, kyn);
}

gantryUpdateVarXsincos (var, ang, kyn)
void *var;
float *ang;
KYN *kyn;
{
   
	register GANTRY_KYN *skyn = (GANTRY_KYN*)kyn->ext;
	register GANTRY_VAR *svar = (GANTRY_VAR*)var;

	svar->d[J1] = ang[J1];
	svar->d[J2] = ang[J2];
	svar->d[J3] = ang[J3];
}
