#include <enableIO.h>

#ifdef i86pc
/*
   These routines support the XXXIOC_ENABLE_IOPORTS icotl 
   which gives a user-level process access to the I/O ports. This
   is done by setting the IOPL level in the EFLAGS register to
   0x3. The only problem is that normally EFLAGS is restored off
   the stack when kernel routines return. That means that EFLAGS
   must be set *before* it is saved by the kernel. The way to do
   this is to temporarily install a patch right at the system
   call entry point -- which can be easily done by temporarily
   changing the call gate entry point for system calls to the
   patch location.

	XXXIOC_ENABLE_IOPORTS should be called twice: 
	first with an argument value of 1, to setup the patch,
	and second with an argument of 0, to execute and remove the patch
*/

void enioInit (EnioRec *enio)
{
	int i;

	for (i=0; i<MAX_CALLGATES; i++)
	 { enio->savedCallGate[i].s_base = 0;
	   enio->savedCallGate[i].s_limacc = 0;
	 }
}

static unsigned long getSegOffset (struct seg_desc *desc)
{
	/* Get the offset stored in a segment descriptor */

	return (((desc->s_limacc & 0xff000000)) +
		((desc->s_limacc & 0xff) << 16) + 
		((desc->s_base & 0xffff0000) >> 16));
}

static unsigned long getGateEntry (struct seg_desc *desc)
{
	/* Returns the entry point stored in a call gate.
	   Not used right now */

	return (((desc->s_limacc & 0xffff0000)) +
		((desc->s_base & 0xffff)));
}

static void genrci_io_patch()
{
	/* Temporary kernel entry point which sets IOPL to 3
	   in the EFLAGS register, before jumping to the normal
	   kernel entry poiint (sys_call()) */

	extern sys_call();

	/* NOTE: EFLAGS register can only be read/set by pushing/popping
	   it on/off tyhe stack */

	asm ("genrci_io_patch_entry:");		/* entry point */
	asm ("pushf");				/* push low word of EFLAGS */
	asm ("orw	$0x3000,(%esp)");	/* set IOPL in pushed value */
	asm ("popf");				/* pop low word of EFLAGS */
	asm ("jmp	sys_call");		/* jump to normal entry point */
}

static void setGateEntry (struct seg_desc *new, struct seg_desc *old, 
		 unsigned long entry)
{
	/* Sets the entry point in a call gate */

	new->s_base = (old->s_base & 0xffff0000) | (entry & 0xffff);
	new->s_limacc = (old->s_limacc & 0xffff) | (entry & 0xffff0000);
}

static unsigned short getLDTReg()
{
	/* Gets the current local descriptor table register.
	   Not used right now */

	asm ("sldt %ax");
}

unsigned short enioGetEflags()
{
	/* Get the low 16 bits of the EFLAGS register. Not used right now */

	asm ("pushf");
	asm ("popw %eax");
}

static int setIOPL()
{
	asm ("pushf");
	asm ("orw $0x3000,(%esp)");
	asm ("popf");	
}

static void doCall (EnioRec *enio, int arg)
{
	struct seg_desc *callgate[MAX_CALLGATES];
	struct seg_desc *ldtdesc;
	struct seg_desc *ldt;	/* local descriptor table */
	struct seg_desc newGate;
	int i, ngates;

	extern void *genrci_io_patch_entry();
	extern cpu_t *cpu_list;

	i = 0;
	if (USER_SCALL & 0x4)
	 { /* call gate is in local descriptor table */
	   ldtdesc = &cpu_list->cpu_m.mcpu_gdt[LDTSEL>>3];
	   ldt = (struct seg_desc*)getSegOffset(ldtdesc);
	   callgate[i++] = &ldt[USER_SCALL>>3];
	 }
	else
	 { /* call gate is in global descriptor table */
	   callgate[i++] = &cpu_list->cpu_m.mcpu_gdt[USER_SCALL>>3];
	 }
#ifdef USER_ALTSCALL
	if (USER_ALTSCALL & 0x4)
	 { /* call gate is in local descriptor table */
	   ldtdesc = &cpu_list->cpu_m.mcpu_gdt[LDTSEL>>3];
	   ldt = (struct seg_desc*)getSegOffset(ldtdesc);
	   callgate[i++] = &ldt[USER_ALTSCALL>>3];
	 }
	else
	 { /* call gate is in global descriptor table */
	   callgate[i++] = &cpu_list->cpu_m.mcpu_gdt[USER_ALTSCALL>>3];
	 }
#endif
	ngates = i;
	
	for (i=0; i<ngates; i++)
	 { if (arg == 0)
	    { enio->savedCallGate[i] = *callgate[i];
	      setGateEntry (&newGate, &enio->savedCallGate[i],
			    (unsigned long)genrci_io_patch_entry);
	      callgate[i]->s_base = newGate.s_base;
	      callgate[i]->s_limacc = newGate.s_limacc;
	    }
	   else
	    { if (enio->savedCallGate[i].s_base != 0)
	       { callgate[i]->s_base = enio->savedCallGate[i].s_base;
		 callgate[i]->s_limacc = enio->savedCallGate[i].s_limacc;
		 enio->savedCallGate[i].s_base = 0;
	       }
	    }
	 }
}

void enioCall0 (EnioRec *enio)
{
	doCall (enio, 0);
}

void enioCall1 (EnioRec *enio)
{
	doCall (enio, 1);
}

#else

void enioInit (EnioRec *enio)
{
}

void enioCall0 (EnioRec *enio)
{
}

void enioCall1 (EnioRec *enio)
{
}

unsigned short enioGetEflags ()
{
	return (0);
}
#endif

